import tantri.dipoles
import numpy
from typing import Sequence


def row_to_dot(input_dict: dict) -> tantri.dipoles.DotPosition:
	r = input_dict["r"]
	if len(r) != 3:
		raise ValueError(
			f"r parameter in input_dict [{input_dict}] does not have length 3"
		)
	label = input_dict["label"]

	return tantri.dipoles.DotPosition(numpy.array(r), label)


def rows_to_dots(
	dot_dict_array: Sequence[dict],
) -> Sequence[tantri.dipoles.DotPosition]:
	return [row_to_dot(input_dict) for input_dict in dot_dict_array]



import logging
from tantri.meta import __version__
from tantri.simple_telegraph import SimpleTelegraphTimeSeries


def get_version():
	return __version__


logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
	"SimpleTelegraphTimeSeries",
	"__version__",
	"get_version",
]



import numpy
from dataclasses import dataclass, asdict
from enum import Enum


# Lazily just separating this from Dipole where there's additional cached stuff, this is just a thing
# we can use as a DTO For dipole info.
@dataclass
class DipoleTO:
	# assumed len 3
	p: numpy.ndarray
	s: numpy.ndarray

	# should be 1/tau up to some pis
	w: float

	def as_dict(self) -> dict:
		return asdict(self)


class Orientation(str, Enum):
	# Enum for orientation, making string for json serialisation purposes
	#
	# Note that htis might not be infinitely extensible?
	# https://stackoverflow.com/questions/75040733/is-there-a-way-to-use-strenum-in-earlier-python-versions
	XY = "XY"
	Z = "Z"
	RANDOM = "RANDOM"


# A description of the parameters needed to generate random dipoles
@dataclass
class DipoleGenerationConfig:
	# note no actual checks anywhere that these are sensibly defined with min less than max etc.
	x_min: float
	x_max: float
	y_min: float
	y_max: float
	z_min: float
	z_max: float

	mag: float

	# these are log_10 of actual value
	w_log_min: float
	w_log_max: float

	orientation: Orientation

	dipole_count: int
	generation_seed: int

	def __post_init__(self):
		# This allows us to transparently set this with a string, while providing early warning of a type error
		self.orientation = Orientation(self.orientation)

	def as_dict(self) -> dict:
		return_dict = asdict(self)

		return_dict["orientation"] = return_dict["orientation"].value

		return return_dict


class DipoleMeasurementType(Enum):
	ELECTRIC_POTENTIAL = 1
	X_ELECTRIC_FIELD = 2


@dataclass(frozen=True)
class DotPosition:
	# assume len 3
	r: numpy.ndarray
	label: str



from tantri.dipoles.types import (
	DipoleTO,
	DotPosition,
	DipoleMeasurementType,
	DipoleGenerationConfig,
)
from tantri.dipoles.time_series import DipoleTimeSeries, WrappedDipole
from tantri.dipoles.generation import make_dipoles

__all__ = [
	"WrappedDipole",
	"DipoleTimeSeries",
	"DipoleTO",
	"DotPosition",
	"DipoleMeasurementType",
	"make_dipoles",
	"DipoleGenerationConfig",
]



import dataclasses
import logging
import math

_logger = logging.getLogger(__name__)

# how many times faster than the max frequency we want to be, bigger is more accurate but 10 is probably fine
DESIRED_THRESHOLD = 10


@dataclasses.dataclass
class SuperSample:
	super_dt: float
	super_sample_ratio: int


def get_supersample(max_frequency: float, dt: float) -> SuperSample:
	# now we want to sample at least 10x faster than max_frequency, otherwise we're going to skew our statistics
	# note that this is why if performance mattered we'd be optimising this to pre-gen our flip times with poisson statistics.
	# so we want (1/dt) > 10 * max_freq
	if DESIRED_THRESHOLD * dt * max_frequency < 1:
		# can return unchanged
		_logger.debug("no supersampling needed")
		return SuperSample(super_dt=dt, super_sample_ratio=1)
	else:
		# else we want a such that a / dt > 10 * max_freq, or a > 10 * dt * max_freq, a = math.ceil(10 * dt * max_freq)
		a = math.ceil(DESIRED_THRESHOLD * dt * max_frequency)
		_logger.debug(
			f"max_frequency {max_frequency} and delta_t {dt} needs a ratio of {a}"
		)
		ret_val = SuperSample(super_dt=dt / a, super_sample_ratio=a)
		_logger.debug(ret_val)
		return ret_val



import numpy
import json


class NumpyEncoder(json.JSONEncoder):
	"""
	Stolen from https://stackoverflow.com/questions/26646362/numpy-array-is-not-json-serializable
	"""

	def default(self, obj):
		if isinstance(obj, numpy.ndarray):
			return obj.tolist()
		return json.JSONEncoder.default(self, obj)
